import datetime
import random

import dateutil.tz
from nose.tools import eq_
import pytz
import six

from taskw.utils import (
    convert_dict_to_override_args,
    decode_task,
    encode_task,
    encode_task_experimental,
    DATE_FORMAT,
    clean_ctrl_chars,
)

TASK = {'description': "task 2 http://www.google.com/",
        'entry': "1325011643",
        'project': "work",
        'due': "1359090000",
        'start': "1326079920", 'status': "pending",
        'uuid': "c1c431ea-f0dc-4683-9a20-e64fcfa65fd1"}


TASK_LEADING_WS = TASK.copy()
TASK_LEADING_WS.update({'description': "      task 3"})


def shuffled(l):
    new = list(l)
    random.shuffle(new)
    return new


class TestUtils(object):

    def test_no_side_effects(self):
        orig = TASK.copy()
        decode_task(encode_task(TASK))
        eq_(orig, TASK)

    def test_with_escaped_quotes(self):
        expected = {'this': r'has a "quote" in it.'}
        line = r'[this:"has a \"quote\" in it."]'
        r = decode_task(line)
        eq_(r, expected)

    def test_with_escaped_quotes_roundtrip(self):
        expected = {'this': r'has a "quote" in it.'}
        line = r'[this:"has a \"quote\" in it."]'
        r = decode_task(encode_task(decode_task(line)))
        eq_(r, expected)

    def test_with_escaped_quotes_full(self):
        line = r'[this:"has a \"quote\" in it."]'
        r = encode_task(decode_task(line))
        eq_(r, r)

    def test_with_backticks(self):
        expected = {'this': r'has a fucking `backtick` in it'}
        line = r'[this:"has a fucking `backtick` in it"]'
        r = decode_task(line)
        eq_(r, expected)
        r = decode_task(encode_task(decode_task(line)))
        eq_(r, expected)

    def test_with_backslashes(self):
        expected = {'andthis': r'has a fucking \backslash in it'}
        line = r'[andthis:"has a fucking \\backslash in it"]'
        r = decode_task(line)
        eq_(r, expected)
        r = decode_task(encode_task(decode_task(line)))
        eq_(r, expected)

    def test_with_unicode(self):
        expected = {
            six.text_type('andthis'): (
                six.text_type('has a fucking \\backslash in it')
            )
        }
        line = r'[andthis:"has a fucking \\backslash in it"]'
        r = decode_task(line)
        eq_(r, expected)
        r = decode_task(encode_task(decode_task(line)))
        eq_(r, expected)

    def test_decode(self):
        r = decode_task(encode_task(TASK))
        eq_(r, TASK)

    def test_decode_leading_whitespace_in_value(self):
        r = decode_task(encode_task(TASK_LEADING_WS))
        eq_(r, TASK_LEADING_WS)

    def test_composition(self):
        eq_(TASK, decode_task(encode_task(TASK)))

    def test_double_composition(self):
        eq_(TASK, decode_task(encode_task(decode_task(encode_task(TASK)))))

    def test_ordering(self):
        task1 = dict(shuffled(TASK.items()))
        task2 = dict(shuffled(TASK.items()))
        eq_(encode_task(task1), encode_task(task2))

    def test_taskwarrior_null_encoding_bug_workaround(self):
        task = {
            'priority': ''
        }
        actual_encoded = encode_task_experimental(task)[0]
        expected_encoded = "priority:"

        eq_(
            actual_encoded,
            expected_encoded
        )

    def test_encodes_dates(self):
        arbitrary_date = datetime.date(2014, 3, 2)
        task = {
            'arbitrary_field': arbitrary_date
        }

        actual_encoded_task = encode_task_experimental(task)
        expected_encoded_task = encode_task_experimental(
            {
                'arbitrary_field': arbitrary_date.strftime(DATE_FORMAT)
            }
        )

        eq_(
            actual_encoded_task,
            expected_encoded_task,
        )

    def test_encodes_naive_datetimes(self):
        arbitrary_naive_datetime = datetime.datetime.now()
        task = {
            'arbitrary_field': arbitrary_naive_datetime
        }

        actual_encoded_task = encode_task_experimental(task)
        expected_encoded_task = encode_task_experimental(
            {
                'arbitrary_field': (
                    arbitrary_naive_datetime
                    .replace(tzinfo=dateutil.tz.tzlocal())
                    .astimezone(pytz.utc).strftime(DATE_FORMAT)
                )
            }
        )

        eq_(
            actual_encoded_task,
            expected_encoded_task,
        )

    def test_encodes_zoned_datetimes(self):
        arbitrary_timezone = pytz.timezone('America/Los_Angeles')
        arbitrary_zoned_datetime = datetime.datetime.now().replace(
            tzinfo=arbitrary_timezone
        )
        task = {
            'arbitrary_field': arbitrary_zoned_datetime
        }

        actual_encoded_task = encode_task_experimental(task)
        expected_encoded_task = encode_task_experimental(
            {
                'arbitrary_field': (
                    arbitrary_zoned_datetime
                    .astimezone(pytz.utc).strftime(DATE_FORMAT)
                )
            }
        )

        eq_(
            actual_encoded_task,
            expected_encoded_task,
        )

    def test_convert_dict_to_override_args(self):
        overrides = {
            'one': {
                'two': 1,
                'three': {
                    'alpha': 'a'
                },
                'four': 'lorem ipsum',
            },
            'two': {
            }
        }

        expected_overrides = [
            'rc.one.two=1',
            'rc.one.three.alpha=a',
            'rc.one.four="lorem ipsum"',
        ]
        actual_overrides = convert_dict_to_override_args(overrides)

        eq_(set(actual_overrides), set(expected_overrides))


class TestCleanExecArg(object):
    def test_clean_null(self):
        eq_(b"", clean_ctrl_chars(b"\x00"))

    def test_all_ctrl_chars(self):
        """ Test that most (but not all) control characters are removed """
        # input = bytes(range(0x20))
        input = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f' # For python 2 compatibility
        eq_(b"\t\n\v\f\r", clean_ctrl_chars(input))



from .base import Field


class ChoiceField(Field):
    def __init__(
        self,
        choices=None,
        nullable=False,
        case_sensitive=False,
        **kwargs
    ):
        self._choices = choices if choices else []
        self._case_sensitive = case_sensitive
        super(ChoiceField, self).__init__(**kwargs)

    def is_valid_choice(self, value):
        if value is None and value not in self._choices:
            return False
        if value is None and value in self._choices:
            return True
        if self._case_sensitive and value in self._choices:
            return True
        elif (
            not self._case_sensitive
            and value.upper() in [v.upper() for v in self._choices if v]
        ):
            return True
        elif self._case_sensitive and value in self._choices:
            return True
        return False

    def serialize(self, value):
        if not self.is_valid_choice(value):
            raise ValueError(
                "'%s' is not a valid choice; choices: %s" % (
                    value,
                    self._choices,
                )
            )
        return value



import sys


class TaskwarriorError(Exception):
    def __init__(self, command, stderr, stdout, code):
        self.command = command
        self.stderr = stderr.strip()
        self.stdout = stdout.strip()
        self.code = code
        super(TaskwarriorError, self).__init__(self.stderr)

    def __unicode__(self):
        return "%r #%s; stderr:\"%s\"; stdout:\"%s\"" % (
            self.command,
            self.code,
            self.stderr,
            self.stdout,
        )

    def __str__(self):
        return self.__unicode__().encode(sys.getdefaultencoding(), 'replace')



import nose
from nose.tools import eq_, ok_, raises
import os
import shutil
import tempfile
import datetime
import dateutil.tz

from taskw import TaskWarriorDirect, TaskWarriorShellout


TASK = {'description': "task 2 http://www.google.com/",
        'entry': "1325011643",
        'project': "work",
        'start': "1326079920", 'status': "pending",
        'uuid': "c1c431ea-f0dc-4683-9a20-e64fcfa65fd1"}


class _BaseTestDB(object):
    def setup(self):

        # Sometimes the 'task' command line tool is not installed.
        if self.should_skip():
            raise nose.SkipTest(
                "%r unsupported on this system" % (self.class_to_test)
            )

        # Create some temporary config stuff
        fd, fname = tempfile.mkstemp(prefix='taskw-testsrc')
        dname = tempfile.mkdtemp(prefix='taskw-tests-data')

        with open(fname, 'w') as f:
            f.writelines([
                'data.location=%s\n' % dname,
                'uda.somestring.label=Testing String\n',
                'uda.somestring.type=string\n',
                'uda.somedate.label=Testing Date\n',
                'uda.somedate.type=date\n',
                'uda.somenumber.label=Testing Number\n',
                'uda.somenumber.type=numeric\n',
            ])

        # Create empty .data files
        for piece in ['completed', 'pending', 'undo']:
            with open(os.path.sep.join([dname, piece + '.data']), 'w'):
                pass

        # Save names for .tearDown()
        self.fname, self.dname = fname, dname

        # Create the taskwarrior db object that each test will use.
        self.tw = self.class_to_test(config_filename=fname)

        # Create a taskwarrior db object for tests where marshal=True,
        # but only for TaskwarriorShellout
        if self.class_to_test == TaskWarriorShellout:
            self.tw_marshal = self.class_to_test(config_filename=fname,
                                                 marshal=True)

    def tearDown(self):
        os.remove(self.fname)
        shutil.rmtree(self.dname)

    def test_has_two_categories(self):
        tasks = self.tw.load_tasks()
        eq_(len(tasks), 2)
        ok_('pending' in tasks)
        ok_('completed' in tasks)

    def test_empty_db(self):
        tasks = self.tw.load_tasks()
        eq_(len(sum(tasks.values(), [])), 0)

    def test_add(self):
        self.tw.task_add("foobar")
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)

    def test_unchanging_load_tasks(self):
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)

    @raises(KeyError)
    def test_completion_raising_unspecified(self):
        self.tw.task_done()

    def test_completing_task_by_id_unspecified(self):
        self.tw.task_add("foobar")
        self.tw.task_done(id=1)
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)
        eq_(len(tasks['completed']), 1)
        eq_(len(sum(tasks.values(), [])), 1)
        ok_(tasks['completed'][0]['end'] is not None)
        eq_(tasks['completed'][0]['status'], 'completed')

    def test_completing_task_by_id_specified(self):
        self.tw.task_add("foobar")
        self.tw.task_done(id=1)
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)
        eq_(len(tasks['completed']), 1)
        eq_(len(sum(tasks.values(), [])), 1)
        eq_(tasks['completed'][0]['status'], 'completed')

    def test_completing_task_by_id_retrieved(self):
        task = self.tw.task_add("foobar")
        self.tw.task_done(id=task['id'])
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)
        eq_(len(tasks['completed']), 1)
        eq_(len(sum(tasks.values(), [])), 1)
        eq_(tasks['completed'][0]['status'], 'completed')

    def test_completing_task_by_uuid(self):
        self.tw.task_add("foobar")
        uuid = self.tw.load_tasks()['pending'][0]['uuid']
        self.tw.task_done(uuid=uuid)
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)
        eq_(len(tasks['completed']), 1)
        eq_(len(sum(tasks.values(), [])), 1)
        eq_(tasks['completed'][0]['status'], 'completed')

    @raises(KeyError)
    def test_get_task_mismatch(self):
        self.tw.task_add("foobar")
        self.tw.task_add("bazbar")
        uuid = self.tw.load_tasks()['pending'][0]['uuid']
        self.tw.get_task(id=2, uuid=uuid)  # which one?

    def test_updating_task(self):
        self.tw.task_add("foobar")

        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)

        task = tasks['pending'][0]
        task["priority"] = "L"
        self.tw.task_update(task)

        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)
        eq_(tasks['pending'][0]['priority'], 'L')

        # For compatibility with the direct and shellout modes.
        # Shellout returns more information.
        try:
            # Shellout mode returns the correct urgency, so,
            # let's just not compare for now.
            del tasks['pending'][0]['urgency']
            del task['urgency']

            # Also, experimental mode returns the id.  So, avoid comparing.
            del tasks['pending'][0]['id']

            # Task 2.2.0 adds a "modified" field, so delete this.
            del tasks['pending'][0]['modified']
        except:
            pass

        # But Task 2.4.0 puts the modified field in earlier
        if 'modified' in task:
            del task['modified']

        eq_(tasks['pending'][0], task)

    @raises(KeyError)
    def test_update_exc(self):
        task = dict(description="lol")
        self.tw.task_update(task)

    def test_add_complicated(self):
        self.tw.task_add(
            "foobar",
            uuid="1234-1234",
            project="some_project"
        )
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)

    def test_add_timestamp(self):
        self.tw.task_add(
            "foobar",
            uuid="1234-1234",
            project="some_project",
            entry="20110101T000000Z",
        )
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)
        eq_(tasks['pending'][0]['entry'], "20110101T000000Z")

    def test_add_datetime(self):
        self.tw.task_add(
            "foobar",
            uuid="1234-1234",
            project="some_project",
            entry=datetime.datetime(2011, 1, 1, tzinfo=dateutil.tz.tzutc()),
        )
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)
        # The exact string we get back is dependent on your current TZ
        # ... we'll just "roughly" test it instead of mocking.
        assert(tasks['pending'][0]['entry'].startswith("20110101T"))

    def test_add_with_uda_string(self):
        self.tw.task_add(
            "foobar",
            somestring="this is a uda",
        )
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)
        task = tasks['pending'][0]

        eq_(task['somestring'], "this is a uda")

    def test_add_with_uda_date(self):
        self.tw.task_add(
            "foobar",
            somedate=datetime.datetime(2011, 1, 1, tzinfo=dateutil.tz.tzutc()),
        )
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)
        task = tasks['pending'][0]

        assert(task['somedate'].startswith("20110101T"))

    @raises(KeyError)
    def test_remove_uda_string(self):
        # Check that a string UDA is removed from a task when its
        # value is set to None
        task = self.tw.task_add(
            "foobar",
            somestring='this is a uda',
        )
        task['somestring'] = None
        id, task = self.tw.task_update(task)
        task['somestring']

    @raises(KeyError)
    def test_remove_uda_date(self):
        # Check that a date UDA is removed from a task when its
        # value is set to None
        task = self.tw.task_add(
            "foobar",
            somedate=datetime.datetime(2011, 1, 1),
        )
        task['somedate'] = None
        id, task = self.tw.task_update(task)
        task['somedate']

    @raises(KeyError)
    def test_remove_uda_numeric(self):
        # Check that a numeric UDA is removed from a task when its
        # value is set to None
        task = self.tw.task_add(
            "foobar",
            somenumber=15,
        )
        task['somenumber'] = None
        id, task = self.tw.task_update(task)
        task['somenumber']

    @raises(ValueError)
    def test_completing_completed_task(self):
        task = self.tw.task_add("foobar")
        self.tw.task_done(uuid=task['uuid'])
        self.tw.task_done(uuid=task['uuid'])

    def test_updating_completed_task(self):
        task = self.tw.task_add("foobar")
        task = self.tw.task_done(uuid=task['uuid'])
        task['priority'] = 'L'
        id, task = self.tw.task_update(task)
        eq_(task['priority'], 'L')

    def test_get_task_completed(self):
        task = self.tw.task_add("foobar")
        task = self.tw.task_done(uuid=task['uuid'])

        id, _task = self.tw.get_task(uuid=task['uuid'])
        eq_(id, None)
        eq_(_task['uuid'], task['uuid'])

    def test_load_task_pending_command(self):
        tasks = self.tw.load_tasks(command='pending')
        eq_(len(tasks), 1)
        ok_('pending' in tasks)

    def test_load_task_completed_command(self):
        tasks = self.tw.load_tasks(command='completed')
        eq_(len(tasks), 1)
        ok_('completed' in tasks)

    @raises(ValueError)
    def test_load_task_with_unknown_command(self):
        tasks = self.tw.load_tasks(command='foobar')

    def test_updating_deleted_task(self):
        task = self.tw.task_add("foobar")
        task = self.tw.task_delete(uuid=task['uuid'])
        task['priority'] = 'L'
        id, task = self.tw.task_update(task)
        eq_(task['priority'], 'L')

    def test_delete(self):
        task = self.tw.task_add("foobar")
        self.tw.task_delete(uuid=task['uuid'])
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)
        # The shellout and direct methods behave differently here
        #eq_(len(tasks['completed']), 1)
        #ok_(not tasks['completed'][0]['end'] is None)
        #eq_(tasks['completed'][0]['status'], 'deleted')

    @raises(ValueError)
    def test_delete_already_deleted(self):
        task = self.tw.task_add("foobar")
        self.tw.task_delete(uuid=task['uuid'])
        self.tw.task_delete(uuid=task['uuid'])

    def test_load_tasks_with_one_each(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_done(uuid=task2['uuid'])
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 1)
        eq_(len(tasks['completed']), 1)

        # For issue #26, I thought this would raise an exception...
        task = self.tw.get_task(description='foobar1')


class TestDBDirect(_BaseTestDB):
    class_to_test = TaskWarriorDirect

    def test_delete_completed(self):
        task = self.tw.task_add("foobar")
        task = self.tw.task_done(uuid=task['uuid'])
        self.tw.task_delete(uuid=task['uuid'])
        tasks = self.tw.load_tasks()
        eq_(len(tasks['pending']), 0)
        eq_(len(tasks['completed']), 1)
        #eq_(tasks['completed'][0]['status'], 'deleted')

    def should_skip(self):
        return False


class TestDBShellout(_BaseTestDB):
    class_to_test = TaskWarriorShellout

    def should_skip(self):
        """ If 'task' is not installed, we can't run these tests. """
        return not TaskWarriorShellout.can_use()

    def test_filtering_simple(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        tasks = self.tw.filter_tasks({
            'description.contains': 'foobar2',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 2)

    def test_filtering_brace(self):
        task1 = self.tw.task_add("[foobar1]")
        task2 = self.tw.task_add("[foobar2]")
        tasks = self.tw.filter_tasks({
            'description.contains': '[foobar2]',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 2)

    def test_filtering_quote(self):
        task1 = self.tw.task_add("[foobar1]")
        task2 = self.tw.task_add("\"foobar2\"")
        tasks = self.tw.filter_tasks({
            'description.contains': '"foobar2"',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 2)

    def test_filtering_plus(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foobar+")
        tasks = self.tw.filter_tasks({
            'description.contains': 'foobar+',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 3)

    def test_filtering_minus(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foobar-")
        tasks = self.tw.filter_tasks({
            'description.contains': 'foobar-',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 3)

    def test_filtering_colon(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foobar:")
        tasks = self.tw.filter_tasks({
            'description.contains': 'foobar:',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 3)

    def test_filtering_qmark(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foo?bar")
        tasks = self.tw.filter_tasks({
                'description.contains': 'oo?ba',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 2)

    def test_filtering_qmark_not_contains(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foo?bar")
        tasks = self.tw.filter_tasks({
                'description': 'foo?bar',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 2)

    def test_filtering_semicolon(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foo;bar")
        tasks = self.tw.filter_tasks({
            'description.contains': 'foo;bar',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 3)

    def test_filtering_question_mark(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foo?bar")
        tasks = self.tw.filter_tasks({
            'description.contains': 'foo?bar',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 3)

    def test_filtering_slash(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foo/bar")
        tasks = self.tw.filter_tasks({
            'description.contains': 'foo/bar',
        })
        eq_(len(tasks), 1)
        eq_(tasks[0]['id'], 3)

    #def test_filtering_double_dash(self):
    #    task1 = self.tw.task_add("foobar1")
    #    task2 = self.tw.task_add("foobar2")
    #    task2 = self.tw.task_add("foo -- bar")
    #    tasks = self.tw.filter_tasks({
    #        'description.contains': 'foo -- bar',
    #    })
    #    eq_(len(tasks), 1)
    #    eq_(tasks[0]['id'], 3)
    #    eq_(tasks[0]['description'], 'foo -- bar')

    def test_filtering_logic_disjunction(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foobar3")
        tasks = self.tw.filter_tasks({
            'or': [
                ('description.has', 'foobar1'),
                ('description.has', 'foobar3'),
            ]
        })
        eq_(len(tasks), 2)
        eq_(tasks[0]['id'], 1)
        eq_(tasks[1]['id'], 3)

    def test_filtering_logic_conjunction(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foobar3")
        tasks = self.tw.filter_tasks({
            'and': [
                ('description.has', 'foobar1'),
                ('description.has', 'foobar3'),
            ]
        })
        eq_(len(tasks), 0)

    def test_filtering_logic_conjunction_junction_whats_your_function(self):
        task1 = self.tw.task_add("foobar1")
        task2 = self.tw.task_add("foobar2")
        task2 = self.tw.task_add("foobar3")
        tasks = self.tw.filter_tasks({
            'and': [
                ('description', 'foobar1'),
            ],
            'or': [
                ('status', 'pending'),
                ('status', 'waiting'),
            ]
        })
        eq_(len(tasks), 1)

    def test_annotation_escaping(self):
        original = {'description': 're-opening the issue'}

        self.tw.task_add('foobar')
        task = self.tw.load_tasks()['pending'][0]
        task['annotations'] = [original]
        self.tw.task_update(task)

        task = self.tw.load_tasks()['pending'][0]
        self.tw.task_update(task)

        eq_(len(task['annotations']), 1)
        eq_(task['annotations'][0]['description'], original['description'])

    @raises(KeyError)
    def test_remove_uda_string_marshal(self):
        # Check that a string UDA is removed from a task when its
        # value is set to None
        task = self.tw_marshal.task_add(
            "foobar",
            somestring='this is a uda',
        )
        task['somestring'] = None
        id, task = self.tw_marshal.task_update(task)
        task['somestring']

    @raises(KeyError)
    def test_remove_uda_date_marshal(self):
        # Check that a date UDA is removed from a task when its
        # value is set to None
        task = self.tw_marshal.task_add(
            "foobar",
            somedate=datetime.datetime(2011, 1, 1),
        )
        task['somedate'] = None
        id, task = self.tw_marshal.task_update(task)
        task['somedate']

    @raises(KeyError)
    def test_remove_uda_numeric_marshal(self):
        # Check that a numeric UDA is removed from a task when its
        # value is set to None
        task = self.tw_marshal.task_add(
            "foobar",
            somenumber=15,
        )
        task['somenumber'] = None
        id, task = self.tw_marshal.task_update(task)
        task['somenumber']

    def test_add_and_retrieve_uda_string_url(self):
        arbitrary_url = "http://www.someurl.com/1084/"

        self.tw.config_overrides['uda'] = {
            'someurl': {
                'label': 'Some URL',
                'type': 'string'
            }
        }
        self.tw.task_add(
            "foobar",
            someurl=arbitrary_url,
        )
        results = self.tw.filter_tasks({
            'someurl.is': arbitrary_url
        })
        eq_(len(results), 1)
        task = results[0]
        eq_(task['someurl'], arbitrary_url)

    def test_add_and_retrieve_uda_string_url_in_parens(self):
        arbitrary_url = "http://www.someurl.com/1084/"

        self.tw.config_overrides['uda'] = {
            'someurl': {
                'label': 'Some URL',
                'type': 'string'
            }
        }
        self.tw.task_add(
            "foobar",
            someurl=arbitrary_url,
        )
        results = self.tw.filter_tasks({
            'and': [
                ('someurl.is', arbitrary_url),
            ],
        })
        eq_(len(results), 1)
        task = results[0]
        eq_(task['someurl'], arbitrary_url)



import datetime

import dateutil
from dateutil.parser import parse
import pytz

from taskw.utils import DATE_FORMAT

from .base import Field


class DateField(Field):
    def deserialize(self, value):
        if not value:
            return value
        value = parse(value)
        if not value.tzinfo:
            value = value.replace(tzinfo=pytz.utc)
        return value

    def serialize(self, value):
        if isinstance(value, datetime.datetime):
            if not value.tzinfo:
                #  Dates not having timezone information should be
                #  assumed to be in local time
                value = value.replace(tzinfo=dateutil.tz.tzlocal())
            #  All times should be converted to UTC before serializing
            value = value.astimezone(pytz.utc).strftime(DATE_FORMAT)
        elif isinstance(value, datetime.date):
            value = value.strftime(DATE_FORMAT)
        return value



import copy
import sys

import six


class Field(object):
    def __init__(self, label=None, read_only=False):
        self._label = label
        self._read_only = read_only
        super(Field, self).__init__()

    @property
    def read_only(self):
        return self._read_only

    @property
    def label(self):
        return self._label

    def deserialize(self, value):
        return value

    def serialize(self, value):
        return value

    def __str__(self):
        if sys.version_info >= (3, ):
            return self.label
        return self.__unicode__().encode(sys.getdefaultencoding(), 'replace')

    def __unicode__(self):
        return self.label

    def __repr__(self):
        return "<{cls} '{label}'>".format(
            cls=six.text_type(self.__class__.__name__),
            label=six.text_type(self) if self._label else '(No Label)',
        )

    def __eq__(self, other):
        if self.label != other.label:
            return False
        if self.read_only != other.read_only:
            return False
        if self.__class__ != other.__class__:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Dirtyable(object):
    """ Superclass for all objects implementing trackability."""

    def __init__(self, value=None):
        self._original_value = copy.deepcopy(value)
        super(Dirtyable, self).__init__(value)

    def get_changes(self, keep=False):
        if self._original_value == self:
            return {}
        result = (self._original_value, self)
        if not keep:
            self._original_value = copy.deepcopy(self)
        return result


class DirtyableList(Dirtyable, list):
    pass


class DirtyableDict(Dirtyable, dict):
    pass



import nose
from nose.tools import eq_
import os
import shutil
import tempfile

from taskw import TaskWarriorShellout


TASK = {'description': "task 2 http://www.google.com/",
        'entry': "1325011643",
        'project': "work",
        'start': "1326079920", 'status': "pending",
        'uuid': "c1c431ea-f0dc-4683-9a20-e64fcfa65fd1"}


class TestRecursibe(object):
    def setup(self):
        if not TaskWarriorShellout.can_use():
            # Sometimes the 'task' command line tool is not installed.
            raise nose.SkipTest("taskwarrior not installed")

        # Create some temporary config stuff
        fd, fname = tempfile.mkstemp(prefix='taskw-testsrc')
        dname = tempfile.mkdtemp(prefix='taskw-tests-data')

        with open(fname, 'w') as f:
            f.writelines([
                'data.location=%s\n' % dname,
                'uda.somestring.label=Testing String\n',
                'uda.somestring.type=string\n',
                'uda.somedate.label=Testing Date\n',
                'uda.somedate.type=date\n',
                'uda.somenumber.label=Testing Number\n',
                'uda.somenumber.type=numeric\n',
            ])

        # Create empty .data files
        for piece in ['completed', 'pending', 'undo']:
            with open(os.path.sep.join([dname, piece + '.data']), 'w'):
                pass

        # Save names for .tearDown()
        self.fname, self.dname = fname, dname

        # Create the taskwarrior db object that each test will use.
        self.tw = TaskWarriorShellout(config_filename=fname, marshal=True)

    def tearDown(self):
        os.remove(self.fname)
        shutil.rmtree(self.dname)

    def test_set_dep_on_one_uuid(self):
        task1 = self.tw.task_add('task1')
        task2 = self.tw.task_add('task2', depends=[task1['uuid']])
        eq_(task2['depends'][0], task1['uuid'])

    def test_set_dep_on_two_uuid(self):
        task1 = self.tw.task_add('task1')
        task2 = self.tw.task_add('task2')
        depends = [task1['uuid'], task2['uuid']]
        task3 = self.tw.task_add('task3', depends=depends)
        eq_(task3['depends'], [task1['uuid'], task2['uuid']])



""" Code to interact with taskwarrior

This module contains an abstract base class and two different implementations
for interacting with taskwarrior:  TaskWarriorDirect and TaskWarriorShellout.

If it is determined that there is a binary 'task' on the system and that it is
of a sufficiently advanced version, then TaskWarriorShellout will be made the
default TaskWarrior class.  If not, then the default TaskWarrior class will
fall back to the older TaskWarriorDirect implementation.

"""
import abc
import codecs
import copy
from distutils.version import LooseVersion
import logging
import os
import time
import uuid
import subprocess
import json
import errno

import kitchen.text.converters

import six
from six import with_metaclass
from six.moves import filter
from six.moves import map

import taskw.utils
from taskw.exceptions import TaskwarriorError
from taskw.task import Task
from taskw.taskrc import TaskRc


logger = logging.getLogger(__name__)


open = lambda fname, mode: codecs.open(fname, mode, "utf-8")

# Location of configuration file: either specified by TASKRC environment
# variable, or ~/.taskrc (default).
TASKRC = os.getenv("TASKRC", "~/.taskrc")


class TaskWarriorBase(with_metaclass(abc.ABCMeta, object)):
    """ The task warrior

    Really though, a python object with methods allowing you to interact
    with a taskwarrior database.
    """

    def __init__(
        self,
        config_filename=TASKRC,
        config_overrides=None,
        marshal=False
    ):
        self.config_filename = config_filename
        self.config = TaskWarriorBase.load_config(config_filename)
        if marshal:
            raise NotImplementedError(
                "You must use TaskWarriorShellout to use 'marshal'"
            )
        if config_overrides:
            raise NotImplementedError(
                "You must use TaskWarriorShellout to use 'config_overrides'"
            )

    def _stub_task(self, description, tags=None, **kw):
        """ Given a description, stub out a task dict. """

        # If whitespace is not removed here, TW will do it when we pass the
        # task to it.
        task = {"description": description.strip()}

        # Allow passing "tags" in as part of kw.
        if 'tags' in kw and tags is None:
            task['tags'] = tags
            del(kw['tags'])

        if tags is not None:
            task['tags'] = tags

        task.update(kw)

        # Only UNIX timestamps are currently supported.
        if 'due' in kw:
            task['due'] = str(task['due'])

        return task

    def _extract_annotations_from_task(self, task):
        """ Removes annotations from a task and returns a list of annotations
        """
        annotations = list()

        if 'annotations' in task:
            existing_annotations = task.pop('annotations')
            for v in existing_annotations:
                if isinstance(v, dict):
                    annotations.append(v['description'])
                else:
                    annotations.append(v)

        for key in list(task.keys()):
            if key.startswith('annotation_'):
                annotations.append(task[key])
                del(task[key])
        return annotations

    @abc.abstractmethod
    def load_tasks(self, command='all'):
        """ Load all tasks.

        Similar to TaskWarrior, a specific command may be specified:

            all       - a list of all issues
            pending   - a list of all pending issues
            completed - a list of all completed issues

        By default, the 'all' command is run.

        >>> w = TaskWarrior()
        >>> tasks = w.load_tasks()
        >>> tasks.keys()
        ['completed', 'pending']
        >>> type(tasks['pending'])
        <type 'list'>
        >>> type(tasks['pending'][0])
        <type 'dict'>
        """

    @abc.abstractmethod
    def task_add(self, description, tags=None, **kw):
        """ Add a new task.

        Takes any of the keywords allowed by taskwarrior like proj or prior.
        """
        pass

    @abc.abstractmethod
    def task_done(self, **kw):
        pass

    @abc.abstractmethod
    def task_delete(self, **kw):
        pass

    @abc.abstractmethod
    def _load_task(self, **kw):
        pass

    @abc.abstractmethod
    def task_update(self, task):
        pass

    @abc.abstractmethod
    def get_task(self, **kw):
        pass

    def filter_by(self, func):
        tasks = self.load_tasks()
        filtered = filter(func, tasks)
        return filtered

    @classmethod
    def load_config(cls, config_filename=TASKRC, overrides=None):
        """ Load ~/.taskrc into a python dict

        >>> config = TaskWarrior.load_config()
        >>> config['data']['location']
        '/home/threebean/.task'
        >>> config['_forcecolor']
        'yes'

        """
        return TaskRc(config_filename, overrides=overrides)

    @abc.abstractmethod
    def task_start(self, **kw):
        pass

    @abc.abstractmethod
    def task_stop(self, **kw):
        pass


class TaskWarriorDirect(TaskWarriorBase):
    """ Interacts with taskwarrior by directly manipulating the ~/.task/ db.

    This is the deprecated implementation and will be phased out in
    time due to taskwarrior's guidelines:  http://bit.ly/16I9VN4

    See https://github.com/ralphbean/taskw/pull/15 for discussion
    and https://github.com/ralphbean/taskw/issues/30 for more.
    """

    def sync(self):
        raise NotImplementedError(
            "You must use TaskWarriorShellout to use 'sync'"
        )

    def load_tasks(self, command='all'):
        def _load_tasks(filename):
            filename = os.path.join(self.config['data']['location'], filename)
            filename = os.path.expanduser(filename)
            with open(filename, 'r') as f:
                lines = f.readlines()

            return list(map(taskw.utils.decode_task, lines))

        return dict(
            (db, _load_tasks(DataFile.filename(db)))
            for db in Command.files(command)
        )

    def get_task(self, **kw):
        line, task = self._load_task(**kw)

        id = None
        # The ID going back only makes sense if the task is pending.
        if Status.is_pending(task['status']):
            id = line

        return id, task

    def _load_task(self, **kw):
        valid_keys = set(['id', 'uuid', 'description'])
        id_keys = valid_keys.intersection(kw.keys())

        if len(id_keys) != 1:
            raise KeyError("Only 1 ID keyword argument may be specified")

        key = list(id_keys)[0]
        if key not in valid_keys:
            raise KeyError("Argument must be one of %r" % valid_keys)

        line = None
        task = dict()

        # If the key is an id, assume the task is pending (completed tasks
        # don't have IDs).
        if key == 'id':
            tasks = self.load_tasks(command=Status.PENDING)
            line = kw[key]

            if len(tasks[Status.PENDING]) >= line:
                task = tasks[Status.PENDING][line - 1]

        else:
            # Search all tasks for the specified key.
            tasks = self.load_tasks(command=Command.ALL)

            matching = list(filter(
                lambda t: t.get(key, None) == kw[key],
                sum(tasks.values(), [])
            ))

            if matching:
                task = matching[0]
                line = tasks[Status.to_file(task['status'])].index(task) + 1

        return line, task

    def task_add(self, description, tags=None, **kw):
        """ Add a new task.

        Takes any of the keywords allowed by taskwarrior like proj or prior.
        """

        task = self._stub_task(description, tags, **kw)

        task['status'] = Status.PENDING

        # TODO -- check only valid keywords

        if not 'entry' in task:
            task['entry'] = str(int(time.time()))

        if not 'uuid' in task:
            task['uuid'] = str(uuid.uuid4())

        id = self._task_add(task, Status.PENDING)
        task['id'] = id
        return task

    def task_done(self, **kw):
        """
        Marks a pending task as done, optionally specifying a completion
        date with the 'end' argument.
        """
        def validate(task):
            if not Status.is_pending(task['status']):
                raise ValueError("Task is not pending.")

        return self._task_change_status(Status.COMPLETED, validate, **kw)

    def task_update(self, task):
        line, _task = self._load_task(uuid=task['uuid'])

        if 'id' in task:
            del task['id']

        # Delete None values (treat them as deleting values)
        # https://github.com/ralphbean/taskw/pull/70
        items = list(task.items())  # listify generator for py3 support.
        for k, v in items:
            if v is None:
                task.pop(k)
                if k in _task:
                    _task.pop(k)

        _task.update(task)
        self._task_replace(line, Status.to_file(task['status']), _task)
        return line, _task

    def task_delete(self, **kw):
        """
        Marks a task as deleted, optionally specifying a completion
        date with the 'end' argument.
        """
        def validate(task):
            if task['status'] == Status.DELETED:
                raise ValueError("Task is already deleted.")

        return self._task_change_status(Status.DELETED, validate, **kw)

    def task_start(self, **kw):
        """ Marks a task as started.  """
        raise NotImplementedError()

    def task_stop(self, **kw):
        """ Marks a task as stopped.  """
        raise NotImplementedError()

    def filter_tasks(self, filter_dict):
        raise NotImplementedError()

    def _task_replace(self, id, category, task):
        def modification(lines):
            lines[id - 1] = taskw.utils.encode_task(task)
            return lines

        # FIXME write to undo.data
        self._apply_modification(id, category, modification)

    def _task_remove(self, id, category):
        def modification(lines):
            del lines[id - 1]
            return lines

        # FIXME write to undo.data
        self._apply_modification(id, category, modification)

    def _apply_modification(self, id, category, modification):
        filename = DataFile.filename(category)
        filename = os.path.join(self.config['data']['location'], filename)
        filename = os.path.expanduser(filename)

        with open(filename, "r") as f:
            lines = f.readlines()

        lines = modification(lines)

        with open(filename, "w") as f:
            f.writelines(lines)

    def _task_add(self, task, category):
        location = self.config['data']['location']
        location = os.path.expanduser(location)
        filename = category + '.data'

        # Append the task
        with open(os.path.join(location, filename), "a") as f:
            f.writelines([taskw.utils.encode_task(task)])

        # FIXME - this gets written when a task is completed.  incorrect.
        # Add to undo.data
        with open(os.path.join(location, 'undo.data'), "a") as f:
            f.write("time %s\n" % str(int(time.time())))
            f.write("new %s" % taskw.utils.encode_task(task))
            f.write("---\n")

        with open(os.path.join(location, filename), "r") as f:
            # The 'id' of this latest added task.
            return len(f.readlines())

    def _task_change_status(self, status, validation, **kw):
        line, task = self._load_task(**kw)
        validation(task)
        original_status = task['status']

        task['status'] = status
        task['end'] = kw.get('end') or str(int(time.time()))

        self._task_add(task, Status.to_file(status))
        self._task_remove(line, Status.to_file(original_status))
        return task


class TaskWarriorShellout(TaskWarriorBase):
    """ Interacts with taskwarrior by invoking shell commands.

    This is currently the supported version and should be considered stable.

    See https://github.com/ralphbean/taskw/pull/15 for discussion
    and https://github.com/ralphbean/taskw/issues/30 for more.
    """
    DEFAULT_CONFIG_OVERRIDES = {
        'json': {
            'array': 'TRUE'
        },
        'verbose': 'nothing',
        'confirmation': 'no',
        'dependency': {
            'confirmation': 'no',
        },
    }

    def __init__(
        self,
        config_filename=TASKRC,
        config_overrides=None,
        marshal=False,
    ):
        super(TaskWarriorShellout, self).__init__(config_filename)
        self.config_overrides = config_overrides if config_overrides else {}
        self._marshal = marshal
        self.config = TaskRc(config_filename, overrides=config_overrides)

        if self.get_version() >= LooseVersion('2.4'):
            self.DEFAULT_CONFIG_OVERRIDES['verbose'] = 'new-uuid'

    def get_configuration_override_args(self):
        config_overrides = self.DEFAULT_CONFIG_OVERRIDES.copy()
        config_overrides.update(self.config_overrides)
        return taskw.utils.convert_dict_to_override_args(config_overrides)

    def _execute(self, *args):
        """ Execute a given taskwarrior command with arguments

        Returns a 2-tuple of stdout and stderr (respectively).

        """
        command = (
            [
                'task',
                'rc:%s' % self.config_filename,
            ]
            + self.get_configuration_override_args()
            + [six.text_type(arg) for arg in args]
        )

        # subprocess is expecting bytestrings only, so nuke unicode if present
        # and remove control characters
        for i in range(len(command)):
            if isinstance(command[i], six.text_type):
                command[i] = (
                    taskw.utils.clean_ctrl_chars(command[i].encode('utf-8')))

        try:
            proc = subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = proc.communicate()
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise OSError("Unable to find the 'task' command-line tool.")
            raise

        if proc.returncode != 0:
            raise TaskwarriorError(command, stderr, stdout, proc.returncode)

        # We should get bytes from the outside world.  Turn those into unicode
        # as soon as we can.
        # Everything going into and coming out of taskwarrior *should* be
        # utf-8, but there are weird edge cases where something totally unusual
        # made it in.. so we need to be able to handle (or at least try to
        # handle) whatever.  Kitchen tries its best.
        try:
            stdout = stdout.decode(self.config.get('encoding', 'utf-8'))
        except UnicodeDecodeError as e:
            stdout = kitchen.text.converters.to_unicode(stdout)
        try:
            stderr = stderr.decode(self.config.get('encoding', 'utf-8'))
        except UnicodeDecodeError as e:
            stderr = kitchen.text.converters.to_unicode(stderr)

        # strip any crazy terminal escape characters like bells, backspaces,
        # and form feeds
        for c in ('\a', '\b', '\f', ''):
            stdout = stdout.replace(c, '?')
            stderr = stderr.replace(c, '?')

        return stdout, stderr

    def _get_json(self, *args):
        return json.loads(self._execute(*args)[0])

    def _get_task_objects(self, *args):
        json = self._get_json(*args)
        if isinstance(json, dict):
            return self._get_task_object(json)
        value = [self._get_task_object(j) for j in json]
        return value

    def _get_task_object(self, obj):
        if self._marshal:
            return Task(obj, udas=self.config.get_udas())
        return obj

    def _stub_task(self, description, tags=None, **kw):
        """ Given a description, stub out a task dict. """

        # If whitespace is not removed here, TW will do it when we pass the
        # task to it.
        task = {"description": description.strip()}

        # Allow passing "tags" in as part of kw.
        if 'tags' in kw and tags is None:
            task['tags'] = tags
            del(kw['tags'])

        if tags is not None:
            task['tags'] = tags

        task.update(kw)

        if self._marshal:
            return Task.from_stub(task, udas=self.config.get_udas())

        return task

    @classmethod
    def can_use(cls):
        """ Returns true if runtime requirements of experimental mode are met
        """
        try:
            return cls.get_version() > LooseVersion('2')
        except OSError:
            # OSError is raised if subprocess.Popen fails to find
            # the executable.
            return False

    @classmethod
    def get_version(cls):
        try:
            taskwarrior_version = subprocess.Popen(
                ['task', '--version'],
                stdout=subprocess.PIPE
            ).communicate()[0]
        except OSError as e:
            if 'No such file or directory' in e:
                raise OSError("Unable to find the 'task' command-line tool.")
            raise
        return LooseVersion(taskwarrior_version.decode())

    def sync(self, init=False):
        if self.get_version() < LooseVersion('2.3'):
            raise UnsupportedVersionException(
                "'sync' requires version 2.3 of taskwarrior or later."
            )
        if init is True:
            self._execute('sync', 'init')
        else:
            self._execute('sync')

    def load_tasks(self, command='all'):
        """ Returns a dictionary of tasks for a list of command."""

        results = dict(
            (db, self._get_task_objects('status:%s' % db, 'export'))
            for db in Command.files(command)
        )

        # 'waiting' tasks are returned separately from 'pending' tasks
        # Here we merge the waiting list back into the pending list.
        if 'pending' in results:
            results['pending'].extend(
                self._get_task_objects('status:waiting', 'export'))

        return results

    def filter_tasks(self, filter_dict):
        """ Return a filtered list of tasks from taskwarrior.

        Filter dict should be a dictionary mapping filter constraints
        with their values.  For example, to return only pending tasks,
        you could use::

            {'status': 'pending'}

        Or, to return tasks that have the word "Abjad" in their description
        that are also pending::

            {
                'status': 'pending',
                'description.contains': 'Abjad',
            }

        Filters can be quite complex, and are documented on Taskwarrior's
        website.

        """
        query_args = taskw.utils.encode_query(filter_dict, self.get_version())
        return self._get_task_objects(
            'export',
            *query_args
        )

    def get_task(self, **kw):
        task = dict()
        task_id = None
        task_id, task = self._load_task(**kw)
        id = None

        # The ID going back only makes sense if the task is pending.
        if 'status' in task:
            if Status.is_pending(task['status']):
                id = task_id

        return id, task

    def _load_task(self, **kwargs):
        if len(kwargs) > 1:
            raise KeyError(
                "Only one keyword argument may be specified"
            )

        search = []
        for key, value in six.iteritems(kwargs):
            if key not in ['id', 'uuid', 'description']:
                search.append(
                    '%s:%s' % (
                        key,
                        value,
                    )
                )
            elif key == 'description' and '(bw)' in value:
                search.append(
                    value[4:]
                )
            else:
                search = [value]

        task = self._get_task_objects('export', *search)

        if task:
            if isinstance(task, list):
                # Multiple items returned from search, return just the 1st
                task = task[0]
            return task['id'], task

        return None, dict()

    def task_add(self, description, tags=None, **kw):
        """ Add a new task.

        Takes any of the keywords allowed by taskwarrior like proj or prior.
        """
        task = self._stub_task(description, tags, **kw)

        # Check if there are annotations, if so remove them from the
        # task and add them after we've added the task.
        annotations = self._extract_annotations_from_task(task)

        # With older versions of taskwarrior, you can specify whatever uuid you
        # want when adding a task.
        if self.get_version() < LooseVersion('2.4'):
            task['uuid'] = str(uuid.uuid4())
        elif 'uuid' in task:
            del task['uuid']

        if self._marshal:
            args = taskw.utils.encode_task_experimental(task.serialized())
        else:
            args = taskw.utils.encode_task_experimental(task)

        stdout, stderr = self._execute('add', *args)

        # However, in 2.4 and later, you cannot specify whatever uuid you want
        # when adding a task.  Instead, you have to specify rc.verbose=new-uuid
        # and then parse the assigned uuid out from stdout.
        if self.get_version() >= LooseVersion('2.4'):
            task['uuid'] = stdout.strip().split()[-1].strip('.')

        id, added_task = self.get_task(uuid=task['uuid'])

        # Check if 'uuid' is in the task we just added.
        if not 'uuid' in added_task:
            raise KeyError(
                'Error encountered while creating task;'
                'STDOUT: %s; STDERR: %s' % (
                    stdout,
                    stderr,
                )
            )

        if annotations and 'uuid' in added_task:
            for annotation in annotations:
                self.task_annotate(added_task, annotation)

        id, added_task = self.get_task(uuid=added_task[six.u('uuid')])
        return added_task

    def task_annotate(self, task, annotation):
        """ Annotates a task. """
        self._execute(
            task['uuid'],
            'annotate',
            '--',
            annotation
        )
        id, annotated_task = self.get_task(uuid=task[six.u('uuid')])
        return annotated_task

    def task_denotate(self, task, annotation):
        """ Removes an annotation from a task. """
        self._execute(
            task['uuid'],
            'denotate',
            '--',
            annotation
        )
        id, denotated_task = self.get_task(uuid=task[six.u('uuid')])
        return denotated_task

    def task_done(self, **kw):
        if not kw:
            raise KeyError('No key was passed.')

        id, task = self.get_task(**kw)

        if not Status.is_pending(task['status']):
            raise ValueError("Task is not pending.")

        self._execute(id, 'done')
        return self.get_task(uuid=task['uuid'])[1]

    def task_update(self, task):
        if 'uuid' not in task:
            raise KeyError('Task must have a UUID.')
        # 'Legacy' causes us to handle this task as if it were an
        # old-style task -- just a standard dictionary
        legacy = True

        if isinstance(task, Task):
            # Let's pre-serialize taskw.task.Task instances
            task_uuid = six.text_type(task['uuid'])
            task = task.serialized_changes(keep=True)
            legacy = False
        else:
            task_uuid = task['uuid']

        id, original_task = self.get_task(uuid=task_uuid)

        if 'id' in task:
            del task['id']

        task_to_modify = copy.deepcopy(task)

        task_to_modify.pop('uuid', None)
        task_to_modify.pop('id', None)

        # Only handle annotation differences if this is an old-style
        # task, or if the task itself says annotations have changed.
        annotations_to_delete = set()
        annotations_to_create = set()
        if legacy or 'annotations' in task_to_modify:
            # Check if there are annotations, if so, look if they are
            # in the existing task, otherwise annotate the task to add them.
            ttm_annotations = taskw.utils.annotation_list_to_comparison_map(
                self._extract_annotations_from_task(task_to_modify)
            )
            original_annotations = (
                taskw.utils.annotation_list_to_comparison_map(
                    self._extract_annotations_from_task(original_task)
                )
            )

            new_annotations = set(ttm_annotations.keys())
            existing_annotations = set(original_annotations.keys())

            annotations_to_delete = existing_annotations - new_annotations
            annotations_to_create = new_annotations - existing_annotations

            if 'annotations' in task_to_modify:
                del task_to_modify['annotations']

        modification = taskw.utils.encode_task_experimental(task_to_modify)
        # Only try to modify the task if there are changes to post here
        # (changes *might* just be in annotations).
        if modification:
            self._execute(task_uuid, 'modify', *modification)

        # If there are no existing annotations, add the new ones
        if legacy or annotations_to_delete or annotations_to_create:
            ttm_annotations.update(original_annotations)
            for annotation_key in annotations_to_create:
                self.task_annotate(
                    original_task,
                    ttm_annotations[annotation_key]
                )
            for annotation_key in annotations_to_delete:
                self.task_denotate(
                    original_task,
                    ttm_annotations[annotation_key]
                )

        return self.get_task(uuid=task_uuid)

    def task_delete(self, **kw):
        """ Marks a task as deleted.  """

        id, task = self.get_task(**kw)

        if task['status'] == Status.DELETED:
            raise ValueError("Task is already deleted.")

        self._execute(id, 'delete')
        return self.get_task(uuid=task['uuid'])[1]

    def task_start(self, **kw):
        """ Marks a task as started.  """

        id, task = self.get_task(**kw)

        self._execute(id, 'start')
        return self.get_task(uuid=task['uuid'])[1]

    def task_stop(self, **kw):
        """ Marks a task as stopped.  """

        id, task = self.get_task(**kw)

        self._execute(id, 'stop')
        return self.get_task(uuid=task['uuid'])[1]

    def task_info(self, **kw):
        id, task = self.get_task(**kw)
        out, err = self._execute(id, 'info')
        if err:
            return err
        return out


class DataFile(object):
    """ Encapsulates data file names. """
    PENDING = 'pending'
    COMPLETED = 'completed'

    @classmethod
    def filename(cls, name):
        return "%s.data" % name


class Command(object):
    """ Encapsulates available commands. """
    PENDING = 'pending'
    COMPLETED = 'completed'
    ALL = 'all'

    @classmethod
    def files(cls, command):
        known_commands = {
            Command.PENDING: [DataFile.PENDING],
            Command.COMPLETED: [DataFile.COMPLETED],
            Command.ALL: [DataFile.PENDING, DataFile.COMPLETED]
        }

        if not command in known_commands:
            raise ValueError(
                "Unknown command, %s. Command must be one of %s." %
                (command, known_commands.keys()))

        return known_commands[command]


class Status(object):
    """ Encapsulates task status values. """
    PENDING = 'pending'
    COMPLETED = 'completed'
    DELETED = 'deleted'
    WAITING = 'waiting'

    @classmethod
    def is_pending(cls, status):
        """ Identifies if the specified status is a 'pending' state. """
        return status == Status.PENDING or status == Status.WAITING

    @classmethod
    def to_file(cls, status):
        """ Returns the file in which this task is stored. """
        return {
            Status.PENDING: DataFile.PENDING,
            Status.WAITING: DataFile.PENDING,
            Status.COMPLETED: DataFile.COMPLETED,
            Status.DELETED: DataFile.COMPLETED
        }[status]


class UnsupportedVersionException(object):
    pass


# It is not really experimental anymore, but we provide this rename for
# backwards compatibility.  It will eventually be removed.
TaskWarriorExperimental = TaskWarriorShellout


# Set a default based on what is available on the system.
if TaskWarriorShellout.can_use():
    TaskWarrior = TaskWarriorShellout
else:
    TaskWarrior = TaskWarriorDirect



from .string import StringField


class DurationField(StringField):
    """ In the future this will handle transforming recurrence patterns.

    See https://github.com/taskwarrior/task/blob/2.3.0/src/Duration.cpp#L41

    """
    pass



from dateutil.parser import parse
import six

from .array import ArrayField
from .base import DirtyableList


class Annotation(six.text_type):
    """ A special type of string that we'll use for storing annotations.

    This is, for all intents and purposes, really just a string, but
    it does allow us to store additional information if we have it -- in
    this application: the annotation's entry date.

    """
    def __new__(self, description, entry=None):
        return six.text_type.__new__(self, description)

    def __init__(self, description, entry=None):
        self._entry = entry
        super(Annotation, self).__init__()

    @property
    def entry(self):
        if self._entry:
            return parse(self._entry)
        return self._entry


class AnnotationArrayField(ArrayField):
    """ A special case of the ArrayField handling idiosyncrasies of Annotations

    Taskwarrior will currently return to you a dictionary of values --
    the annotation's date and description -- for each annotation, but
    given that we cannot create annotations with a date, let's instead
    return something that behaves like a string (but from which you can
    extract an entry date if one exists).

    """
    def deserialize(self, value):
        if not value:
            value = DirtyableList([])

        elements = []
        for annotation in value:
            if isinstance(annotation, dict):
                elements.append(
                    Annotation(
                        annotation['description'],
                        annotation['entry'],
                    )
                )
            else:
                elements.append(Annotation(annotation))

        return super(AnnotationArrayField, self).deserialize(elements)

    def serialize(self, value):
        if not value:
            value = []
        return super(AnnotationArrayField, self).serialize(
            [six.text_type(entry) for entry in value]
        )



import tempfile
import os
import shutil

from taskw.warrior import TaskWarrior


class TestTaskWarrior(object):
    def setUp(self):
        self.taskdata = tempfile.mkdtemp()
        taskrc = os.path.join(os.path.dirname(__file__), 'data/empty.taskrc')
        self.taskwarrior = TaskWarrior(
            config_filename=taskrc,
            config_overrides={'data.location': self.taskdata})
        # Just a sanity check to make sure that after the setup, the list of
        # tasks is empty, otherwise we are probably using the current user's
        # TASKDATA and we should not continue.
        assert self.taskwarrior.load_tasks() == {'completed': [], 'pending': []}

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.taskdata)

    def test_add_task_foobar(self):
        """ Add a task with description 'foobar' and checks that the task is
        indeed created """
        self.taskwarrior.task_add("foobar")
        tasks = self.taskwarrior.load_tasks()
        assert len(tasks['pending']) == 1
        assert tasks['pending'][0]['description'] == 'foobar'

    def test_add_task_null_char(self):
        """ Try adding a task where the description contains a NULL character
        (0x00). This should not fail but instead simply add a task with the
        same description minus the NULL character """
        self.taskwarrior.task_add("foo\x00bar")
        tasks = self.taskwarrior.load_tasks()
        assert len(tasks['pending']) == 1
        assert tasks['pending'][0]['description'] == 'foobar'



import logging

import six

from taskw.utils import encode_replacements_experimental
from .base import Field


logger = logging.getLogger(__name__)


class StringField(Field):
    def deserialize(self, value):
        # If value is None, let's just let it pass through
        if not value:
            return value
        if not isinstance(value, six.string_types):
            value = six.text_type(value)
        for left, right in six.iteritems(encode_replacements_experimental):
            value = value.replace(right, left)
        return value

    def serialize(self, value):
        # If value is None let it pass through
        if not value:
            return value
        if not isinstance(value, six.string_types):
            string_value = six.text_type(value)
            logger.debug(
                "Value %s serialized to string as '%s'",
                repr(value),
                string_value
            )
            value = string_value
        for left, right in six.iteritems(encode_replacements_experimental):
            value = value.replace(left, right)
        return value



import numbers

from .base import Field


class NumericField(Field):
    def serialize(self, value):
        if not isinstance(value, numbers.Number):
            if value:
                raise ValueError("Value must be numeric.")
        return value

    def deserialize(self, value):
        if value is None:
            return value
        try:
            return int(value)
        except ValueError:
            pass
        try:
            return float(value)
        except ValueError:
            pass

        # If we've made it this far, somehow Taskwarrior has
        # a non-numeric value stored in this field; this shouldn't
        # happen, but it's probably inappropriate to blow up.
        return value



""" Various utilties """
from __future__ import print_function

import datetime
import re
from operator import itemgetter

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import dateutil.tz
import pytz
import six

from distutils.version import LooseVersion


DATE_FORMAT = '%Y%m%dT%H%M%SZ'


encode_replacements = OrderedDict([
    ('\\', '\\\\'),
    ('\"', '&dquot;'),
    ('"', '&dquot;'),
    ('[', '&open;'),
    (']', '&close;'),
    ('\n', ' '),
    ('/', '\\/'),
])

encode_replacements_experimental = OrderedDict([
    ('\"', '&dquot;'),
    ('"', '&dquot;'),
    ('[', '&open;'),
    (']', '&close;'),
])

decode_replacements = OrderedDict([
    [v, k] for k, v in encode_replacements.items()
    if k not in ('\n')  # We skip these.
])

logical_replacements = OrderedDict([
    ('?', '\\?'),
    ('+', '\\+'),
    ('(', '\\('),
    (')', '\\)'),
    ('[', '\\['),
    (']', '\\]'),
    ('{', '\\{'),
    ('}', '\\}'),
])


def encode_task_value(key, value, query=False):
    if value is None:
        value = ''
    elif isinstance(value, datetime.datetime):
        if not value.tzinfo:
            #  Dates not having timezone information should be
            #  assumed to be in local time
            value = value.replace(tzinfo=dateutil.tz.tzlocal())
        #  All times should be converted to UTC before serializing
        value = value.astimezone(pytz.utc).strftime(DATE_FORMAT)
    elif isinstance(value, datetime.date):
        value = value.strftime(DATE_FORMAT)
    elif isinstance(value, six.string_types):
        if query:
            # In some contexts, parentheses are interpreted for use in
            # logical expressions.  They must *sometimes* be escaped.
            for left, right in six.iteritems(logical_replacements):
                # don't replace '?' if this is an exact match
                if left == '?' and '.' not in key:
                    continue
                value = value.replace(left, right)
        else:
            for unsafe, safe in six.iteritems(
                encode_replacements_experimental
            ):
                value = value.replace(unsafe, safe)
    else:
        value = str(value)
    return value


def encode_query(value, version, query=True):
    args = []

    if isinstance(value, dict):
        value = six.iteritems(value)

    for k, v in value:
        if isinstance(v, list):
            args.append(
                "( %s )" % (" %s " % k).join([
                    encode_query([item], version, query=False)[0] for item in v
                ])
            )
        else:
            if k.endswith(".is") and version >= LooseVersion('2.4'):
                args.append(
                    '%s == "%s"' % (
                        k[:-3],
                        encode_task_value(k, v, query=query)
                    )
                )
            else:
                args.append(
                    '%s:%s' % (
                        k,
                        encode_task_value(k, v, query=query)
                    )
                )

    return args


def clean_task(task):
    """ Clean a task by replacing any dangerous characters """
    return task


def encode_task_experimental(task):
    """ Convert a dict-like task to its string representation
        Used for adding a task via `task add`
    """
    # First, clean the task:
    task = task.copy()
    if 'tags' in task:
        task['tags'] = ','.join(task['tags'])
    for k in task:
        task[k] = encode_task_value(k, task[k])

    # Then, format it as a string
    return [
        "%s:\"%s\"" % (k, v) if v else "%s:" % (k, )
        for k, v in sorted(task.items(), key=itemgetter(0))
    ]


def encode_task(task):
    """ Convert a dict-like task to its string representation """
    # First, clean the task:
    task = task.copy()
    if 'tags' in task:
        task['tags'] = ','.join(task['tags'])
    for k in task:
        for unsafe, safe in six.iteritems(encode_replacements):
            if isinstance(task[k], six.string_types):
                task[k] = task[k].replace(unsafe, safe)

        if isinstance(task[k], datetime.datetime):
            task[k] = task[k].strftime("%Y%m%dT%M%H%SZ")

    # Then, format it as a string
    return "[%s]\n" % " ".join([
        "%s:\"%s\"" % (k, v)
        for k, v in sorted(task.items(), key=itemgetter(0))
    ])


def decode_task(line):
    """ Parse a single record (task) from a task database file.

    I don't understand why they don't just use JSON or YAML.  But
    that's okay.

    >>> decode_task('[description:"Make a python API for taskwarrior"]')
    {'description': 'Make a python API for taskwarrior'}

    """

    task = {}
    for key, value in re.findall(r'(\w+):"(.*?)(?<!\\)"', line):
        value = value.replace('\\"', '"')  # unescape quotes
        task[key] = value
        for unsafe, safe in six.iteritems(decode_replacements):
            task[key] = task[key].replace(unsafe, safe)
    if 'tags' in task:
        task['tags'] = task['tags'].split(',')
    return task


def make_annotation_comparable(annotation):
    """ Make an annotation comparable.

    Some transformations occur internally when storing a message in
    Taskwarrior.  Let's flatten those out.

    """
    return re.sub(
        r'[\W_]',
        '',
        annotation
    )


def get_annotation_value(annotation):
    """ Can either be a dictionary, or a string. """
    if isinstance(annotation, dict):
        return annotation['description']
    return annotation


def annotation_exists_in_list(authoritative, new):
    comparable_annotations = []
    for item in authoritative:
        if not item:
            continue
        annotation = get_annotation_value(item)
        comparable_annotations.append(
            make_annotation_comparable(annotation)
        )
    return make_annotation_comparable(new) in comparable_annotations


def merge_annotations(left, right):
    for annotation in right:
        if not annotation_exists_in_list(left, annotation):
            left.append(right)

    return left


def annotation_list_to_comparison_map(annotations):
    mapping = {}
    for annotation in annotations:
        comparable = make_annotation_comparable(annotation)
        mapping[comparable] = annotation
    return mapping


def convert_dict_to_override_args(config, prefix=''):
    """ Converts a dictionary of override arguments into CLI arguments.

    * Converts leaf nodes into dot paths of key names leading to the leaf
      node.
    * Does not include paths to leaf nodes not being non-dictionary type.

    See `taskw.test.test_utils.TestUtils.test_convert_dict_to_override_args`
    for details.

    """
    args = []
    for k, v in six.iteritems(config):
        if isinstance(v, dict):
            args.extend(
                convert_dict_to_override_args(
                    v,
                    prefix='.'.join([
                        prefix,
                        k,
                    ]) if prefix else k
                )
            )
        else:
            v = six.text_type(v)
            left = 'rc' + (('.' + prefix) if prefix else '') + '.' + k
            right = v if ' ' not in v else '"%s"' % v
            args.append('='.join([left, right]))
    return args


CTRLCHAR = re.compile(b"[\x00-\x08\x0e-\x1f]")

def clean_ctrl_chars(s):
    """ Clean string removing most (but not all) control characters """
    return CTRLCHAR.sub(b'', s)



from taskw.warrior import (
    TaskWarrior,
    TaskWarriorDirect,
    TaskWarriorShellout,
    TaskWarriorExperimental,
)
from taskw.utils import clean_task, encode_task, decode_task
from taskw.utils import encode_task_experimental

__all__ = [
    TaskWarrior,
    TaskWarriorShellout,
    TaskWarriorDirect,
    TaskWarriorExperimental,  # This is deprecated.  Use TaskWarriorShellout
    clean_task,
    encode_task,
    decode_task,
    encode_task_experimental,
]



import os
import sys

from taskw.warrior import TaskWarrior
from taskw.taskrc import TaskRc
from taskw.fields import NumericField, ChoiceField


if sys.version_info >= (2, 7):
    from unittest import TestCase
else:
    from unittest2 import TestCase


class TestBasicLoading(TestCase):
    def setUp(self):
        self.path_to_taskrc = os.path.join(
            os.path.dirname(__file__),
            'data/default.taskrc',
        )

    def test_load_config(self):
        expected = {
            'data': {
                'location': '~/.task'
            },
            'alpha': {
                'one': 'yes',
                'two': '2',
            },
            'beta': {
                'one': 'FALSE',
            },
            'gamma': {
                'one': 'TRUE',
            },
            'uda': {
                'a': {
                    'type': 'numeric',
                    'label': 'Alpha',
                },
                'b': {
                    'type': 'string',
                    'label': 'Beta',
                    'values': 'Strontium-90,Hydrogen-3',
                }
            }
        }
        config = TaskWarrior.load_config(self.path_to_taskrc)
        self.assertEqual(config, expected)


class TestTaskRc(TestCase):
    def setUp(self):
        self.path_to_taskrc = os.path.join(
            os.path.dirname(__file__),
            'data/default.taskrc',
        )
        self.taskrc = TaskRc(self.path_to_taskrc)

    def test_taskrc_parsing(self):
        expected_config = {
            'data': {
                'location': '~/.task'
            },
            'alpha': {
                'one': 'yes',
                'two': '2',
            },
            'beta': {
                'one': 'FALSE',
            },
            'gamma': {
                'one': 'TRUE',
            },
            'uda': {
                'a': {
                    'type': 'numeric',
                    'label': 'Alpha',
                },
                'b': {
                    'type': 'string',
                    'label': 'Beta',
                    'values': 'Strontium-90,Hydrogen-3',
                }
            }
        }

        self.assertEqual(self.taskrc, expected_config)

    def test_get_udas(self):
        expected_udas = {
            'a': NumericField(label='Alpha'),
            'b': ChoiceField(
                label='Beta',
                choices=['Strontium-90', 'Hydrogen-3'],
            ),
        }
        actual_udas = self.taskrc.get_udas()

        self.assertEqual(actual_udas, expected_udas)

    def test_config_overrides(self):
        overrides = {
            'uda': {
                'd': {
                    'type': 'string',
                    'label': 'Delta',
                }
            },
            'alpha': {
                'two': '3',
            }
        }

        taskrc = TaskRc(self.path_to_taskrc, overrides=overrides)

        expected_config = {
            'data': {
                'location': '~/.task'
            },
            'alpha': {
                'one': 'yes',
                'two': '3',
            },
            'beta': {
                'one': 'FALSE',
            },
            'gamma': {
                'one': 'TRUE',
            },
            'uda': {
                'a': {
                    'type': 'numeric',
                    'label': 'Alpha',
                },
                'b': {
                    'type': 'string',
                    'label': 'Beta',
                    'values': 'Strontium-90,Hydrogen-3',
                },
                'd': {
                    'type': 'string',
                    'label': 'Delta',
                }
            }
        }

        self.assertEqual(taskrc, expected_config)



from .base import DirtyableList, Field


class ArrayField(Field):
    def deserialize(self, value):
        if not value:
            value = DirtyableList([])
        return DirtyableList(value)

    def serialize(self, value):
        if not value:
            value = []
        if not hasattr(value, '__iter__'):
            raise ValueError("Value must be list or tuple.")
        return value



import copy
import datetime
import sys
import uuid

import pytz
import six

from taskw.task import Task

if sys.version_info >= (2, 7):
    from unittest import TestCase
else:
    from unittest2 import TestCase


class TestTaskDirtyability(TestCase):
    def setUp(self):
        self.task = Task({
            'uuid': str(uuid.uuid4()),
            'description': 'Something important',
            'due': (
                datetime.datetime.now().replace(tzinfo=pytz.UTC)
                + datetime.timedelta(hours=1)
            ).strftime('%Y%m%dT%H%M%SZ'),
            'tags': ['one', 'two', 'three'],
        })

    def test_append_when_absent(self):
        self.task['annotations'] = []
        self.task['annotations'].append('awesome')
        self.assertEqual(self.task['annotations'], ['awesome'])

    def test_append_when_absent_but_with_tags(self):
        self.task = Task({'uuid': str(uuid.uuid4()), 'description': 'Test'})
        self.task['tags'] = []
        self.task['tags'].append('awesome')
        self.assertEqual(self.task['tags'], ['awesome'])

    def test_marks_date_changed(self):
        original_due_date = self.task['due']
        new_due_date = datetime.datetime.now().replace(tzinfo=pytz.UTC)
        self.task['due'] = new_due_date

        expected_changes = {'due': (original_due_date, new_due_date)}
        actual_changes = self.task.get_changes()

        self.assertEqual(list(six.iterkeys(actual_changes)), ['due'])

        # Use assertAlmostEqual to allow for millisecond loss when
        # converting to string in setUp
        self.assertAlmostEqual(
            expected_changes['due'][0],
            actual_changes['due'][0],
            delta=datetime.timedelta(seconds=1)
        )
        self.assertAlmostEqual(
            expected_changes['due'][1],
            actual_changes['due'][1],
            delta=datetime.timedelta(seconds=1)
        )

    def test_marks_tags_changed(self):
        original_tags = copy.deepcopy(self.task['tags'])
        new_tag = 'alpha'
        self.task['tags'].append(new_tag)

        expected_tags = copy.deepcopy(original_tags)
        expected_tags.append(new_tag)

        expected_changes = {'tags': [original_tags, expected_tags]}
        actual_changes = self.task.get_changes()

        self.assertEqual(actual_changes, expected_changes)

    def test_does_not_mark_unchanged(self):
        self.task['description'] = self.task['description']

        expected_changes = {}
        actual_changes = self.task.get_changes()

        self.assertEqual(actual_changes, expected_changes)

    def test_does_not_allow_changing_id(self):
        with self.assertRaises(ValueError):
            self.task['id'] = 10


class TestTaskMarshalling(TestCase):
    def test_serialize(self):
        arbitrary_serialized_data = {
            'depends': ','.join([
                str(uuid.uuid4()),
                str(uuid.uuid4()),
            ]),
            'description': '&open;Something important',
            'due': (
                datetime.datetime.now().replace(tzinfo=pytz.UTC)
                + datetime.timedelta(hours=1)
            ).strftime('%Y%m%dT%H%M%SZ'),
            'tags': ['one', 'two', 'three'],
            'urgency': 10,
            'uuid': str(uuid.uuid4()),
        }
        task = Task(arbitrary_serialized_data)
        expected_result = arbitrary_serialized_data
        actual_result = task.serialized()

        self.assertEqual(actual_result, expected_result)

    def test_deserialize(self):
        arbitrary_depends_uuids = [uuid.uuid4(), uuid.uuid4()]
        arbitrary_description = '[one'
        arbitrary_due_date = (
            datetime.datetime.now().replace(tzinfo=pytz.UTC)
            + datetime.timedelta(hours=1)
        )
        arbitrary_tags = ['one', 'two', ]
        arbitrary_urgency = 10
        arbitrary_uuid = uuid.uuid4()

        serialized = {
            'depends': ','.join([str(u) for u in arbitrary_depends_uuids]),
            'description': arbitrary_description.replace('[', '&open;'),
            'due': arbitrary_due_date.strftime('%Y%m%dT%H%M%SZ'),
            'tags': arbitrary_tags,
            'urgency': arbitrary_urgency,
            'uuid': str(arbitrary_uuid)
        }
        task = Task(serialized)

        self.assertEqual(task['depends'], arbitrary_depends_uuids)
        self.assertEqual(task['description'], arbitrary_description)
        # Loss of milliseconds when converting to string
        self.assertAlmostEqual(
            task['due'],
            arbitrary_due_date,
            delta=datetime.timedelta(seconds=1)
        )
        self.assertEqual(task['tags'], arbitrary_tags)
        self.assertEqual(task['urgency'], arbitrary_urgency)
        self.assertEqual(task['uuid'], arbitrary_uuid)

    def test_composition(self):
        arbitrary_serialized_data = {
            'depends': ','.join([
                str(uuid.uuid4()),
                str(uuid.uuid4()),
            ]),
            'description': '&open;Something important',
            'due': (
                datetime.datetime.now().replace(tzinfo=pytz.UTC)
                + datetime.timedelta(hours=1)
            ).strftime('%Y%m%dT%H%M%SZ'),
            'tags': ['one', 'two', 'three'],
            'urgency': 10,
            'uuid': str(uuid.uuid4()),
        }
        task = Task(arbitrary_serialized_data)
        expected_result = arbitrary_serialized_data

        after_composition = Task(
            Task(
                Task(
                    arbitrary_serialized_data
                ).serialized()
            ).serialized()
        ).serialized()

        self.assertEqual(after_composition, expected_result)



from __future__ import absolute_import

import uuid

from .base import Field


class UUIDField(Field):
    def deserialize(self, value):
        if not value:
            return value
        return uuid.UUID(value)

    def serialize(self, value):
        if isinstance(value, uuid.UUID):
            value = str(value)
        else:
            # Just to make sure it's a UUID
            uuid.UUID(value)
        return value



from __future__ import absolute_import

from distutils.version import LooseVersion

import uuid

from .base import DirtyableList, Field


class CommaSeparatedUUIDField(Field):
    version = LooseVersion('2.4')

    def deserialize(self, value):
        if not value:
            return DirtyableList([])

        # In task-2.5, this moved from a comma-separated string to a real list.
        # here we allow a migration path where old splitable strings are
        # handled as well as newschool lists.
        if hasattr(value, 'split'):
            value = value.split(',')

        return DirtyableList([uuid.UUID(v) for v in value])

    def serialize(self, value):
        if not value:
            value = []

        if not hasattr(value, '__iter__'):
            raise ValueError("Value must be list or tuple, not %r." % value)

        if self.version < LooseVersion('2.5'):
            return ','.join([str(v) for v in value])
        else:
            # We never hit this second code branch now.  taskwarrior changed
            # API slightly in version 2.5, but we're just going to go with
            # backwards compatibility for now.
            # Some day we should switch wholesale to the new path.
            return [str(v) for v in value]



import os
import multiprocessing
import sys
from setuptools import setup, find_packages
import uuid

f = open('README.rst')
long_description = f.read().strip()
long_description = long_description.split('split here', 1)[1]
f.close()

REQUIREMENTS_FILES = {
    'test': 'test_requirements.txt',
    'install': 'requirements.txt',
}
REQUIREMENTS = {}
for category, filename in REQUIREMENTS_FILES.items():
    requirements_path = os.path.join(
        os.path.dirname(__file__),
        filename
    )
    try:
        from pip.req import parse_requirements
        requirements = [
            str(req.req) for req in parse_requirements(
                requirements_path,
                session=uuid.uuid1()
            )
        ]
    except ImportError:
        requirements = []
        with open(requirements_path, 'r') as in_:
            requirements = [
                req for req in in_.readlines()
                if not req.startswith('-')
                and not req.startswith('#')
            ]
    REQUIREMENTS[category] = requirements

if sys.version_info < (2, 7):
    REQUIREMENTS['test'].append('unittest2')
    REQUIREMENTS['install'].append('ordereddict')

setup(name='taskw',
      version='1.2.0',
      description="Python bindings for your taskwarrior database",
      long_description=long_description,
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.2",
          "Programming Language :: Python :: 3.3",
          "License :: OSI Approved :: GNU General Public License (GPL)",
          "Intended Audience :: Developers",
      ],
      keywords='taskwarrior task',
      author='Ralph Bean',
      author_email='ralph.bean@gmail.com',
      url='http://github.com/ralphbean/taskw',
      license='GPLv3+',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=REQUIREMENTS['install'],
      test_suite='nose.collector',
      tests_require=REQUIREMENTS['test'],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



import codecs
import logging
import os

from taskw.fields import (
    ChoiceField,
    DateField,
    DurationField,
    NumericField,
    StringField
)


logger = logging.getLogger(__name__)


def sanitize(line):
    comment_position = line.find('#')
    if comment_position < 0:
        return line.strip()
    return line[:comment_position].strip()


class TaskRc(dict):
    """ Access the user's taskRC using a dictionary-like interface.

    There is a downside, though:

    Unfortunately, collapsing our configuration into a dict has a
    jarring limitation -- we can't store both of the following
    simultaneously:

    * color = on
    * color.header = something

    In this module, we err on the side of storing subkeys rather than the
    actual value in situations where both are necessary.

    Please forgive me.

    """

    UDA_TYPE_MAP = {
        'date': DateField,
        'duration': DurationField,
        'numeric': NumericField,
        'string': StringField,
    }

    def __init__(self, path=None, overrides=None):
        self.overrides = overrides if overrides else {}
        if path:
            self.path = os.path.normpath(
                os.path.expanduser(
                    path
                )
            )
            config = self._read(self.path)
        else:
            self.path = None
            config = {}
        super(TaskRc, self).__init__(config)

    def _add_to_tree(self, config, key, value):
        key_parts = key.split('.')
        cursor = config
        for part in key_parts[0:-1]:
            if part not in cursor:
                cursor[part] = {}
            # See class docstring -- we can't store both a value and
            # a dict in the same place.
            if not isinstance(cursor[part], dict):
                cursor[part] = {}
            cursor = cursor[part]
        cursor[key_parts[-1]] = value
        return config

    def _merge_trees(self, left, right):
        if left is None:
            left = {}

        for key, value in right.items():
            # See class docstring -- we can't store both a value and
            # a dict in the same place.
            if not isinstance(left, dict):
                left = {}
            if isinstance(value, dict):
                left[key] = self._merge_trees(left.get(key), value)
            else:
                left[key] = value

        return left

    def _read(self, path):
        config = {}
        with codecs.open(path, 'r', 'utf8') as config_file:
            for raw_line in config_file.readlines():
                line = sanitize(raw_line)
                if not line:
                    continue
                if line.startswith('include '):
                    try:
                        left, right = line.split(' ')
                        config = self._merge_trees(
                            config,
                            TaskRc(right.strip())
                        )
                    except ValueError:
                        logger.exception(
                            "Error encountered while adding TaskRc at "
                            "'%s' (from TaskRc file at '%s')",
                            right.strip(),
                            self.path
                        )
                else:
                    try:
                        left, right = line.split('=')
                        key = left.strip()
                        value = right.strip()
                        config = self._add_to_tree(config, key, value)
                    except ValueError:
                        logger.exception(
                            "Error encountered while processing configuration "
                            "setting '%s' (from TaskRc file at '%s')",
                            line,
                            self.path,
                        )

        return self._merge_trees(config, self.overrides)

    def __delitem__(self, *args):
        raise TypeError('TaskRc objects are immutable')

    def __setitem__(self, item, value):
        raise TypeError('TaskRc objects are immutable')

    def update(self, value):
        raise TypeError('TaskRc objects are immutable')

    def get_udas(self):
        raw_udas = self.get('uda', {})
        udas = {}

        for k, v in raw_udas.items():
            tw_type = v.get('type', '')
            label = v.get('label', None)
            choices = v.get('values', None)

            kwargs = {}
            cls = self.UDA_TYPE_MAP.get(tw_type, StringField)
            if choices:
                cls = ChoiceField
                kwargs['choices'] = choices.split(',')
            if label:
                kwargs['label'] = label

            udas[k] = cls(**kwargs)

        return udas

    def __unicode__(self):
        return 'TaskRc file at {path}'.format(
            path=self.path
        )

    def __str__(self):
        return self.__unicode__().encode('utf-8', 'REPLACE')



import datetime
import uuid
import sys

from dateutil.tz import tzlocal
from pytz import UTC, timezone
import six

from taskw import fields
from taskw.fields.annotationarray import Annotation

if sys.version_info >= (2, 7):
    from unittest import TestCase
else:
    from unittest2 import TestCase


class TestAnnotationArrayField(TestCase):
    def setUp(self):
        self.field = fields.AnnotationArrayField()

    def test_serialize_none(self):
        actual_result = self.field.serialize(None)
        expected_result = []

        self.assertEqual(actual_result, expected_result)

    def test_serialize_annotations_into_strings(self):
        value = [
            Annotation("something", "20240101T010101Z"),
            Annotation("something else")
        ]

        expected_serialized = ["something", "something else"]
        actual_serialized = self.field.serialize(value)

        self.assertEqual(actual_serialized, expected_serialized)
        for entry in actual_serialized:
            self.assertTrue(
                isinstance(entry, six.text_type)
            )

    def test_deserialize_fully_formed_entries_to_stringey_things(self):
        # Note that this test is *identical* in conditions and actions
        # to the below, but we are asserting that we can extract treat
        # the returned entries just as if they were strings.
        value = [
            {
                'description': 'Coddingtonbear\'s birthday',
                'entry': '19840302T000000Z'
            },
            {
                'description': 'Coddingtonbear\'s partner\'s birthday',
                'entry': '19850711T000000Z',
            }
        ]

        expected_results = [
            value[0]['description'],
            value[1]['description'],
        ]
        actual_results = self.field.deserialize(value)

        self.assertEqual(expected_results, actual_results)

    def test_deserialize_fully_formed_entries_to_annotation_objects(self):
        # Note that this test is *identical* in conditions and actions
        # to the above, but we are asserting that we can extract a little
        # bit more information from the returned objects.
        value = [
            {
                'description': 'Coddingtonbear\'s birthday',
                'entry': '19840302T000000Z'
            },
            {
                'description': 'Coddingtonbear\'s partner\'s birthday',
                'entry': '19850711T000000Z',
            }
        ]

        expected_results = [
            Annotation(value[0]['description'], value[0]['entry']),
            Annotation(value[1]['description'], value[1]['entry']),
        ]
        actual_results = self.field.deserialize(value)

        self.assertEqual(expected_results, actual_results)
        self.assertEqual(expected_results[0].entry.year, 1984)
        self.assertEqual(expected_results[1].entry.year, 1985)


class TestArrayField(TestCase):
    def setUp(self):
        self.field = fields.ArrayField()

    def test_serialize_none(self):
        actual_result = self.field.serialize(None)
        expected_result = []

        self.assertEqual(actual_result, expected_result)

    def test_serialize_not_ok(self):
        arbitrary_inappropriate_value = 1

        with self.assertRaises(ValueError):
            self.field.serialize(arbitrary_inappropriate_value)

    def test_serialize_ok(self):
        arbitrary_list = [1, 2, 3]

        actual_result = self.field.serialize(arbitrary_list)
        expected_result = arbitrary_list

        self.assertEqual(actual_result, expected_result)


class TestChoiceField(TestCase):
    def test_serialize_ok(self):
        field = fields.ChoiceField(
            choices=['A', 'B', 'C'],
            nullable=False,
            case_sensitive=True,
        )
        acceptable_value = 'A'

        actual_value = field.serialize(acceptable_value)
        expected_value = acceptable_value

        self.assertEqual(actual_value, expected_value)

    def test_serialize_wrong_case(self):
        field = fields.ChoiceField(
            choices=['A', 'B', 'C'],
            nullable=False,
            case_sensitive=True,
        )
        unacceptable_wrong_case = 'a'

        with self.assertRaises(ValueError):
            field.serialize(unacceptable_wrong_case)

    def test_serialize_none_unacceptable(self):
        field = fields.ChoiceField(
            choices=['A', 'B', 'C'],
            nullable=False,
            case_sensitive=True,
        )
        unacceptable_none_value = None

        with self.assertRaises(ValueError):
            field.serialize(unacceptable_none_value)

    def test_serialize_none_acceptable(self):
        field = fields.ChoiceField(
            choices=[None, 'A', 'B', 'C'],
            case_sensitive=True,
        )
        acceptable_none_value = None

        actual_value = field.serialize(acceptable_none_value)
        expected_value = acceptable_none_value

        self.assertEqual(actual_value, expected_value)

    def test_serialize_case_insensitive(self):
        field = fields.ChoiceField(
            choices=['A', 'B', 'C'],
            nullable=False,
            case_sensitive=False,
        )
        value_wrong_case = 'a'

        actual_value = field.serialize(value_wrong_case)
        expected_value = value_wrong_case

        self.assertEqual(actual_value, expected_value)


class TestCommaSeparatedUUIDField(TestCase):
    def setUp(self):
        self.field = fields.CommaSeparatedUUIDField()

    def test_serialize_single_uuid(self):
        single_uuid = [uuid.uuid4()]

        actual_value = self.field.serialize(single_uuid)
        expected_value = str(single_uuid[0])

        self.assertEqual(actual_value, expected_value)

    def test_serialize_mulitple_values(self):
        many_uuids = [
            uuid.uuid4(),
            uuid.uuid4(),
            uuid.uuid4(),
        ]

        actual_value = self.field.serialize(many_uuids)
        expected_value = ','.join([str(u) for u in many_uuids])

        self.assertEqual(actual_value, expected_value)

    def test_deserialize_uuid_string(self):
        arbitrary_uuids = [uuid.uuid4(), uuid.uuid4()]
        uuid_strings = ','.join([str(u) for u in arbitrary_uuids])

        actual_value = self.field.deserialize(uuid_strings)
        expected_value = arbitrary_uuids

        self.assertEqual(actual_value, expected_value)

    def test_deserialize_uuid_string_undashed(self):
        arbitrary_uuids = [uuid.uuid4(), uuid.uuid4()]
        uuid_strings = ','.join([u.hex for u in arbitrary_uuids])

        actual_value = self.field.deserialize(uuid_strings)
        expected_value = arbitrary_uuids

        self.assertEqual(actual_value, expected_value)


class TestDateField(TestCase):
    def setUp(self):
        self.field = fields.DateField()

    def test_deserialize_none(self):
        actual_value = self.field.deserialize(None)
        expected_value = None

        self.assertEqual(actual_value, expected_value)

    def test_deserialize_naive(self):
        arbitrary_year = 2014
        arbitrary_month = 3
        arbitrary_day = 2
        arbitrary_hour = 9
        arbitrary_minute = 10
        arbitrary_second = 3

        naive_date_string = (
            '{year}-{month}-{day}T{hour}:{minute}:{second}'.format(
                year=arbitrary_year,
                month=arbitrary_month,
                day=arbitrary_day,
                hour=arbitrary_hour,
                minute=arbitrary_minute,
                second=arbitrary_second,
            )
        )

        actual_value = self.field.deserialize(naive_date_string)
        expected_value = datetime.datetime(
            arbitrary_year,
            arbitrary_month,
            arbitrary_day,
            arbitrary_hour,
            arbitrary_minute,
            arbitrary_second,
            tzinfo=UTC
        )

        self.assertEqual(actual_value, expected_value)

    def test_deserialize_nonnaive(self):
        arbitrary_year = 2014
        arbitrary_month = 3
        arbitrary_day = 2
        arbitrary_hour = 9
        arbitrary_minute = 10
        arbitrary_second = 3

        nonnaive_date_string = (
            '{year}-{month}-{day}T{hour}:{minute}:{second}Z'.format(
                year=arbitrary_year,
                month=arbitrary_month,
                day=arbitrary_day,
                hour=arbitrary_hour,
                minute=arbitrary_minute,
                second=arbitrary_second,
            )
        )

        actual_value = self.field.deserialize(nonnaive_date_string)
        expected_value = datetime.datetime(
            arbitrary_year,
            arbitrary_month,
            arbitrary_day,
            arbitrary_hour,
            arbitrary_minute,
            arbitrary_second,
            tzinfo=UTC
        )

        self.assertEqual(actual_value, expected_value)

    def test_serialize_none(self):
        actual_value = self.field.serialize(None)
        expected_value = None

        self.assertEqual(actual_value, expected_value)

    def test_serialize_naive(self):
        arbitrary_year = 2014
        arbitrary_month = 3
        arbitrary_day = 2
        arbitrary_hour = 9
        arbitrary_minute = 10
        arbitrary_second = 3

        arbitrary_date = datetime.datetime(
            arbitrary_year,
            arbitrary_month,
            arbitrary_day,
            arbitrary_hour,
            arbitrary_minute,
            arbitrary_second,
        )

        actual_value = self.field.serialize(arbitrary_date)
        expected_value = UTC.normalize(
            arbitrary_date.replace(tzinfo=tzlocal())
        ).strftime('%Y%m%dT%H%M%SZ')

        self.assertEqual(actual_value, expected_value)

    def test_serialize_nonnaive(self):
        arbitrary_year = 2014
        arbitrary_month = 3
        arbitrary_day = 2
        arbitrary_hour = 9
        arbitrary_minute = 10
        arbitrary_second = 3

        arbitrary_date = datetime.datetime(
            arbitrary_year,
            arbitrary_month,
            arbitrary_day,
            arbitrary_hour,
            arbitrary_minute,
            arbitrary_second,
            tzinfo=timezone('America/Los_Angeles')
        )

        actual_value = self.field.serialize(arbitrary_date)
        expected_value = UTC.normalize(
            arbitrary_date
        ).strftime('%Y%m%dT%H%M%SZ')

        self.assertEqual(actual_value, expected_value)


class TestNumericField(TestCase):
    def setUp(self):
        self.field = fields.NumericField()

    def test_deserialize_float(self):
        arbitrary_float = '214.8'

        actual_value = self.field.deserialize(arbitrary_float)
        expected_value = 214.8

        self.assertEqual(actual_value, expected_value)

    def test_deserialize_integer(self):
        arbitrary_integer = '214'

        actual_value = self.field.deserialize(arbitrary_integer)
        expected_value = 214

        self.assertEqual(actual_value, expected_value)

    def test_deserialize_and_lo_and_behold_it_wasnt_numeric(self):
        arbitrary_string = 'alpha'

        actual_value = self.field.deserialize(arbitrary_string)
        expected_value = arbitrary_string

        self.assertEqual(actual_value, expected_value)

    def test_numeric_value(self):
        arbitrary_numeric_value = 10

        actual_value = self.field.serialize(arbitrary_numeric_value)
        expected_value = arbitrary_numeric_value

        self.assertEqual(actual_value, expected_value)

    def test_nonnumeric_value(self):
        arbitrary_nonnumeric_value = 'alpha'

        with self.assertRaises(ValueError):
            self.field.serialize(arbitrary_nonnumeric_value)


class TestStringField(TestCase):
    def setUp(self):
        self.field = fields.StringField()

    def test_deserialize_string(self):
        serialized_string = '&open;hello&close;'

        actual_value = self.field.deserialize(serialized_string)
        expected_value = '[hello]'

        self.assertEqual(actual_value, expected_value)

    def test_serialize_string(self):
        unserialized_string = '[hello]'

        actual_value = self.field.serialize(unserialized_string)
        expected_value = '&open;hello&close;'

        self.assertEqual(actual_value, expected_value)


class TestUUIDField(TestCase):
    def setUp(self):
        self.field = fields.UUIDField()

    def test_serialize(self):
        arbitrary_uuid = uuid.uuid4()

        actual_result = self.field.serialize(arbitrary_uuid)
        expected_result = str(arbitrary_uuid)

        self.assertEqual(actual_result, expected_result)

    def test_deserialize_dashed(self):
        arbitrary_uuid = uuid.uuid4()
        dashed_uuid = str(arbitrary_uuid)

        actual_result = self.field.deserialize(dashed_uuid)
        expected_result = arbitrary_uuid

        self.assertEqual(actual_result, expected_result)

    def test_deserialize_undashed(self):
        arbitrary_uuid = uuid.uuid4()
        dashed_uuid = arbitrary_uuid.hex

        actual_result = self.field.deserialize(dashed_uuid)
        expected_result = arbitrary_uuid

        self.assertEqual(actual_result, expected_result)



import logging

import six

from taskw.fields import (
    AnnotationArrayField,
    ArrayField,
    ChoiceField,
    CommaSeparatedUUIDField,
    DateField,
    DurationField,
    Field,
    NumericField,
    StringField,
    UUIDField,
)
from taskw.fields.base import Dirtyable, DirtyableList, DirtyableDict

# Sentinel value for not specifying a default
UNSPECIFIED = object()


logger = logging.getLogger(__name__)


class Task(dict):
    FIELDS = {
        'annotations': AnnotationArrayField(label='Annotations'),
        'depends': CommaSeparatedUUIDField(label='Depends Upon'),
        'description': StringField(label='Description'),
        'due': DateField(label='Due'),
        'end': DateField(label='Ended'),
        'entry': DateField(label='Entered'),
        'id': NumericField(label='ID', read_only=True),
        'imask': NumericField(label='Imask', read_only=True),
        'mask': StringField(label='Mask', read_only=True),
        'modified': DateField(label='Modified'),
        'parent': StringField(label='Parent'),
        'priority': ChoiceField(
            choices=[None, 'H', 'M', 'L', ],
            case_sensitive=False,
            label='Priority'
        ),
        'project': StringField(label='Project'),
        'recur': DurationField(label='Recurrence'),
        'scheduled': DateField(label='Scheduled'),
        'start': DateField(label='Started'),
        'status': ChoiceField(
            choices=[
                'pending',
                'completed',
                'deleted',
                'waiting',
                'recurring',
            ],
            case_sensitive=False,
            label='Status',
        ),
        'tags': ArrayField(label='Tags'),
        'until': DateField(label='Until'),
        'urgency': NumericField(label='Urgency', read_only=True),
        'uuid': UUIDField(label='UUID'),
        'wait': DateField(label='Wait'),
    }

    def __init__(self, data, udas=None):
        udas = udas or {}
        self._fields = self.FIELDS.copy()
        self._fields.update(udas)
        self._changes = []

        processed = {}
        for k, v in six.iteritems(data):
            processed[k] = self._deserialize(k, v, self._fields)

        super(Task, self).__init__(processed)

    @classmethod
    def from_stub(cls, data, udas=None):
        """ Create a Task from an already deserialized dict. """

        udas = udas or {}
        fields = cls.FIELDS.copy()
        fields.update(udas)

        processed = {}
        for k, v in six.iteritems(data):
            processed[k] = cls._serialize(k, v, fields)

        return cls(processed, udas)

    @classmethod
    def _get_converter_for_field(cls, field, default=None, fields=None):
        fields = fields or {}
        converter = fields.get(field, None)
        if not converter:
            return default if default else Field()
        return converter

    @classmethod
    def _deserialize(cls, key, value, fields):
        """ Marshal incoming data into Python objects."""
        converter = cls._get_converter_for_field(key, None, fields)
        return converter.deserialize(value)

    @classmethod
    def _serialize(cls, key, value, fields):
        """ Marshal outgoing data into Taskwarrior's JSON format."""
        converter = cls._get_converter_for_field(key, None, fields)
        return converter.serialize(value)

    def _field_is_writable(self, key):
        converter = self._get_converter_for_field(key, fields=self._fields)
        if converter.read_only:
            return False
        return True

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def _record_change(self, key, from_, to):
        self._changes.append((key, from_, to))

    def get_changes(self, serialized=False, keep=False):
        """ Get a journal of changes that have occurred

        :param `serialized`:
            Return changes in the serialized format used by TaskWarrior.
        :param `keep_changes`:
            By default, the list of changes is reset after running
            ``.get_changes``; set this to `True` if you would like to
            keep the changes recorded following running this command.

        :returns: A dictionary of 2-tuples of changes, where the key is the
            name of the field that has changed, and the value is a 2-tuple
            containing the original value and the final value respectively.

        """
        results = {}

        # Check for explicitly-registered changes
        for k, f, t in self._changes:
            if k not in results:
                results[k] = [f, None]
            results[k][1] = (
                self._serialize(k, t, self._fields)
                if serialized else t
            )

        # Check for changes on subordinate items
        for k, v in six.iteritems(self):
            if isinstance(v, Dirtyable):
                result = v.get_changes(keep=keep)
                if result:
                    if not k in results:
                        results[k] = [result[0], None]
                    results[k][1] = (
                        self._serialize(k, result[1], self._fields)
                        if serialized else result[1]
                    )

        # Clear out recorded changes
        if not keep:
            self._changes = []

        return results

    def update(self, values, force=False):
        """ Update this task dictionary

        :returns: A dictionary mapping field names specified to be updated
            and a boolean value indicating whether the field was changed.

        """
        results = {}
        for k, v in six.iteritems(values):
            results[k] = self.__setitem__(k, v, force=force)
        return results

    def set(self, key, value):
        """ Set a key's value regardless of whether a change is seen."""
        return self.__setitem__(key, value, force=True)

    def serialized(self):
        """ Returns a serialized representation of this task."""
        serialized = {}
        for k, v in six.iteritems(self):
            serialized[k] = self._serialize(k, v, self._fields)
        return serialized

    def serialized_changes(self, keep=False):
        serialized = {}
        for k, v in six.iteritems(self.get_changes(keep=keep)):
            # Here, `v` is a 2-tuple of the field's original value
            # and the field's new value.
            _, to = v
            serialized[k] = self._serialize(k, to, self._fields)
        return serialized

    def __setitem__(self, key, value, force=False):
        if isinstance(value, dict) and not isinstance(value, DirtyableDict):
            value = DirtyableDict(value)
        elif isinstance(value, list) and not isinstance(value, DirtyableList):
            value = DirtyableList(value)

        existing_value = self.get(key)
        if force or value != existing_value:
            if force or existing_value or value:
                # Do not attempt to record changes if both the existing
                # and previous values are Falsy.  We cannot distinguish
                # between `''` and `None` for...reasons.
                self._record_change(
                    key,
                    self.get(key),
                    value,
                )

            # Serialize just to make sure we can; it's better to throw
            # this error early.
            self._serialize(key, value, self._fields)

            # Also make sure we raise an error if this field isn't
            # writable at all.
            if not self._field_is_writable(key):
                raise ValueError("%s is a read-only field", key)

            super(Task, self).__setitem__(key, value)
            return True
        return False



from .base import Field
from .annotationarray import AnnotationArrayField
from .array import ArrayField
from .commaseparateduuid import CommaSeparatedUUIDField
from .choice import ChoiceField
from .date import DateField
from .duration import DurationField
from .numeric import NumericField
from .string import StringField
from .uuid import UUIDField
